@@ -14,7 +14,10 @@ pub fn derive_asn1_read(input: proc_macro::TokenStream) -> proc_macro::TokenStre
     let lifetime_name = add_lifetime_if_none(&mut generics);
     add_bounds(
         &mut generics,
+        all_field_types(&input.data),
         syn::parse_quote!(asn1::Asn1Readable<#lifetime_name>),
+        syn::parse_quote!(asn1::Asn1DefinedByReadable<#lifetime_name, asn1::ObjectIdentifier>),
+        false,
     );
     let (impl_generics, _, where_clause) = generics.split_for_impl();
 
@@ -58,7 +61,13 @@ pub fn derive_asn1_write(input: proc_macro::TokenStream) -> proc_macro::TokenStr
     let mut input = syn::parse_macro_input!(input as syn::DeriveInput);
 
     let name = input.ident;
-    add_bounds(&mut input.generics, syn::parse_quote!(asn1::Asn1Writable));
+    add_bounds(
+        &mut input.generics,
+        all_field_types(&input.data),
+        syn::parse_quote!(asn1::Asn1Writable),
+        syn::parse_quote!(asn1::Asn1DefinedByWritable<asn1::ObjectIdentifier>),
+        true,
+    );
     let (impl_generics, ty_generics, where_clause) = input.generics.split_for_impl();
 
     let expanded = match input.data {
@@ -83,6 +92,12 @@ pub fn derive_asn1_write(input: proc_macro::TokenStream) -> proc_macro::TokenStr
                         #write_block
                     }
                 }
+
+                impl #impl_generics asn1::Asn1Writable for &#name #ty_generics #where_clause {
+                    fn write(&self, w: &mut asn1::Writer) -> asn1::WriteResult {
+                        (*self).write(w)
+                    }
+                }
             }
         }
         _ => unimplemented!("Not supported for unions"),
@@ -260,14 +275,141 @@ fn add_lifetime_if_none(generics: &mut syn::Generics) -> syn::Lifetime {
     generics.lifetimes().next().unwrap().lifetime.clone()
 }
 
-fn add_bounds(generics: &mut syn::Generics, bound: syn::TypeParamBound) {
-    for param in &mut generics.params {
-        if let syn::GenericParam::Type(ref mut type_param) = param {
-            type_param.bounds.push(bound.clone());
+fn all_field_types(data: &syn::Data) -> Vec<(syn::Type, OpType, bool)> {
+    let mut field_types = vec![];
+    match data {
+        syn::Data::Struct(v) => {
+            add_field_types(&mut field_types, &v.fields, None);
         }
+        syn::Data::Enum(v) => {
+            for variant in &v.variants {
+                let (op_type, _) = extract_field_properties(&variant.attrs);
+                add_field_types(&mut field_types, &variant.fields, Some(op_type));
+            }
+        }
+        syn::Data::Union(_) => panic!("Unions not supported"),
+    }
+    field_types
+}
+
+fn add_field_types(
+    field_types: &mut Vec<(syn::Type, OpType, bool)>,
+    fields: &syn::Fields,
+    op_type: Option<OpType>,
+) {
+    match fields {
+        syn::Fields::Named(v) => {
+            for f in &v.named {
+                add_field_type(field_types, f, op_type.clone());
+            }
+        }
+        syn::Fields::Unnamed(v) => {
+            for f in &v.unnamed {
+                add_field_type(field_types, f, op_type.clone());
+            }
+        }
+        syn::Fields::Unit => {}
     }
 }
 
+fn add_field_type(
+    field_types: &mut Vec<(syn::Type, OpType, bool)>,
+    f: &syn::Field,
+    op_type: Option<OpType>,
+) {
+    // If we have an op_type here, it means it came from an enum variant. In
+    // that case, even though it wasn't marked "required", it is for the
+    // purposes of how we're using it.
+    let (op_type, default) = if let Some(OpType::Explicit(mut args)) = op_type {
+        args.required = true;
+        (OpType::Explicit(args), None)
+    } else if let Some(OpType::Implicit(mut args)) = op_type {
+        args.required = true;
+        (OpType::Implicit(args), None)
+    } else {
+        extract_field_properties(&f.attrs)
+    };
+    field_types.push((f.ty.clone(), op_type, default.is_some()));
+}
+
+fn add_bounds(
+    generics: &mut syn::Generics,
+    field_types: Vec<(syn::Type, OpType, bool)>,
+    bound: syn::TypeParamBound,
+    defined_by_bound: syn::TypeParamBound,
+    add_ref: bool,
+) {
+    let where_clause = if field_types.is_empty() {
+        return;
+    } else {
+        generics
+            .where_clause
+            .get_or_insert_with(|| syn::WhereClause {
+                where_token: Default::default(),
+                predicates: syn::punctuated::Punctuated::new(),
+            })
+    };
+
+    for (f, op_type, has_default) in field_types {
+        let (bounded_ty, required_bound) = match (op_type, add_ref) {
+            (OpType::Regular, _) => (f, bound.clone()),
+            (OpType::DefinedBy(_), _) => (f, defined_by_bound.clone()),
+
+            (OpType::Implicit(OpTypeArgs { value, required }), false) => {
+                let ty = if required || has_default {
+                    syn::parse_quote!(asn1::Implicit::<#f, #value>)
+                } else {
+                    syn::parse_quote!(asn1::Implicit::<<#f as asn1::OptionExt>::T, #value>)
+                };
+
+                (ty, bound.clone())
+            }
+            (OpType::Implicit(OpTypeArgs { value, required }), true) => {
+                let ty = if required || has_default {
+                    syn::parse_quote!(for<'asn1_internal> asn1::Implicit::<&'asn1_internal #f, #value>)
+                } else {
+                    syn::parse_quote!(for<'asn1_internal> asn1::Implicit::<&'asn1_internal <#f as asn1::OptionExt>::T, #value>)
+                };
+
+                (ty, bound.clone())
+            }
+
+            (OpType::Explicit(OpTypeArgs { value, required }), false) => {
+                let ty = if required || has_default {
+                    syn::parse_quote!(asn1::Explicit::<#f, #value>)
+                } else {
+                    syn::parse_quote!(asn1::Explicit::<<#f as asn1::OptionExt>::T, #value>)
+                };
+
+                (ty, bound.clone())
+            }
+            (OpType::Explicit(OpTypeArgs { value, required }), true) => {
+                let ty = if required || has_default {
+                    syn::parse_quote!(for<'asn1_internal> asn1::Explicit::<&'asn1_internal #f, #value>)
+                } else {
+                    syn::parse_quote!(for<'asn1_internal> asn1::Explicit::<&'asn1_internal <#f as asn1::OptionExt>::T, #value>)
+                };
+
+                (ty, bound.clone())
+            }
+        };
+
+        where_clause
+            .predicates
+            .push(syn::WherePredicate::Type(syn::PredicateType {
+                lifetimes: None,
+                bounded_ty,
+                colon_token: Default::default(),
+                bounds: {
+                    let mut p = syn::punctuated::Punctuated::new();
+                    p.push(required_bound);
+                    p
+                },
+            }))
+    }
+}
+
+#[derive(Clone)]
 enum OpType {
     Regular,
     Explicit(OpTypeArgs),
@@ -275,6 +417,7 @@ enum OpType {
     DefinedBy(syn::Ident),
 }
 
+#[derive(Clone)]
 struct OpTypeArgs {
     value: proc_macro2::Literal,
     required: bool,
@@ -376,8 +519,9 @@ fn generate_read_element(
         },
     };
     if let Some(default) = default {
+        let f_type = &f.ty;
         read_op = quote::quote! {{
-            asn1::from_optional_default(#read_op, #default.into())#add_error_location?
+            asn1::from_optional_default::<#f_type>(#read_op, #default.into())#add_error_location?
         }};
     }
     read_op
@@ -492,7 +636,7 @@ fn generate_enum_read_block(
             OpType::Explicit(arg) => {
                 let tag = arg.value;
                 read_blocks.push(quote::quote! {
-                    if tlv.tag() == asn1::explicit_tag(#tag) {
+                    if asn1::Explicit::<#ty, #tag>::can_parse(tlv.tag()) {
                         return Ok(#name::#ident(asn1::parse(
                             tlv.full_data(),
                             |p| Ok(p.read_element::<asn1::Explicit<_, #tag>>()#add_error_location?.into_inner())
@@ -500,7 +644,7 @@ fn generate_enum_read_block(
                     }
                 });
                 can_parse_blocks.push(quote::quote! {
-                    if tag == asn1::explicit_tag(#tag) {
+                    if asn1::Explicit::<#ty, #tag>::can_parse(tag) {
                         return true;
                     }
                 });
@@ -508,7 +652,7 @@ fn generate_enum_read_block(
             OpType::Implicit(arg) => {
                 let tag = arg.value;
                 read_blocks.push(quote::quote! {
-                    if tlv.tag() == asn1::implicit_tag(#tag, <#ty as asn1::SimpleAsn1Readable>::TAG) {
+                    if asn1::Implicit::<#ty, #tag>::can_parse(tlv.tag()) {
                         return Ok(#name::#ident(asn1::parse(
                             tlv.full_data(),
                             |p| Ok(p.read_element::<asn1::Implicit<_, #tag>>()#add_error_location?.into_inner())
@@ -516,7 +660,7 @@ fn generate_enum_read_block(
                     }
                 });
                 can_parse_blocks.push(quote::quote! {
-                    if tag == asn1::implicit_tag(#tag, <#ty as asn1::SimpleAsn1Readable>::TAG) {
+                    if asn1::Implicit::<#ty, #tag>::can_parse(tag) {
                         return true;
                     }
                 });
@@ -552,11 +696,13 @@ fn generate_write_element(
             let value = arg.value;
             if arg.required {
                 quote::quote_spanned! {f.span() =>
-                    w.write_explicit_element(#field_read, #value)?;
+                    w.write_element(&asn1::Explicit::<_, #value>::new(#field_read))?;
                 }
             } else {
                 quote::quote_spanned! {f.span() =>
-                    w.write_optional_explicit_element(#field_read, #value)?;
+                    if let Some(v) = #field_read {
+                        w.write_element(&asn1::Explicit::<_, #value>::new(v))?;
+                    }
                 }
             }
         }
@@ -564,11 +710,13 @@ fn generate_write_element(
             let value = arg.value;
             if arg.required {
                 quote::quote_spanned! {f.span() =>
-                    w.write_implicit_element(#field_read, #value)?;
+                    w.write_element(&asn1::Implicit::<_, #value>::new(#field_read))?;
                 }
             } else {
                 quote::quote_spanned! {f.span() =>
-                    w.write_optional_implicit_element(#field_read, #value)?;
+                    if let Some(v) = #field_read {
+                        w.write_element(&asn1::Implicit::<_, #value>::new(v))?;
+                    }
                 }
             }
         }
@@ -657,13 +805,13 @@ fn generate_enum_write_block(name: &syn::Ident, data: &syn::DataEnum) -> proc_ma
             OpType::Explicit(arg) => {
                 let tag = arg.value;
                 quote::quote! {
-                    #name::#ident(value) => w.write_explicit_element(&value, #tag),
+                    #name::#ident(value) => w.write_element(&asn1::Explicit::<_, #tag>::new(value)),
                 }
             }
             OpType::Implicit(arg) => {
                 let tag = arg.value;
                 quote::quote! {
-                    #name::#ident(value) => w.write_implicit_element(&value, #tag),
+                    #name::#ident(value) => w.write_element(&asn1::Implicit::<_, #tag>::new(value)),
                 }
             }
             OpType::DefinedBy(_) => panic!("Can't use #[defined_by] in an Asn1Write on an enum"),
@@ -678,7 +826,7 @@ fn generate_enum_write_block(name: &syn::Ident, data: &syn::DataEnum) -> proc_ma
 
 // TODO: Duplicate of this function in src/object_identifier.rs, can we
 // de-dupe?
-fn _write_base128_int(data: &mut Vec<u8>, n: u32) {
+fn _write_base128_int(data: &mut Vec<u8>, n: u128) {
     if n == 0 {
         data.push(0);
         return;
@@ -709,11 +857,11 @@ pub fn oid(item: proc_macro::TokenStream) -> proc_macro::TokenStream {
     let mut arcs = p_arcs.iter();
 
     let mut der_encoded = vec![];
-    let first = arcs.next().unwrap().base10_parse::<u32>().unwrap();
-    let second = arcs.next().unwrap().base10_parse::<u32>().unwrap();
+    let first = arcs.next().unwrap().base10_parse::<u128>().unwrap();
+    let second = arcs.next().unwrap().base10_parse::<u128>().unwrap();
     _write_base128_int(&mut der_encoded, 40 * first + second);
     for arc in arcs {
-        _write_base128_int(&mut der_encoded, arc.base10_parse().unwrap());
+        _write_base128_int(&mut der_encoded, arc.base10_parse::<u128>().unwrap());
     }
 
     let der_len = der_encoded.len();
